/*
 * Functions to generate anonymous Ada types
 */
#include <sys/types.h>
#include "lowlevel.h"
#include "host.h"
#include "files.h"
#include "hash.h"
#include "il.h"
#include "nodeop.h"
#include "allocate.h"
#include "types.h"
#include "stab.h"
#include "ada_name.h"
#include "anonymous.h"
#include "config.h"

extern int auto_package;
extern file_pos_t yypos;

#undef TYPE_HASH_MAX
#define TYPE_HASH_MAX	256		/* Anonymous type hash table size */

#undef NULL
#define NULL			0

static typeinfo_t *anonymous_types[TYPE_HASH_MAX];
static int anonymous_ord[MAX_UNIQ_FNAMES];

symbol_t *anonymous_function_pointer;

/*
 * Use ordinal for anonymous type names
 */
int
next_anonymous_ord()
{
	int file_ord;

	if (auto_package) {
		file_ord = FILE_ORD(yypos);
		assert(file_ord < num_files());
		return anonymous_ord[file_ord]++;
	}

	return anonymous_ord[0]++;
}

/*
 * don't generate names ending in _t_t
 */
static int
ends_in_t(str)
	char *str;
{
	for (; *str; str++) {
		if (str[0] == '_' && str[1] == 't' && str[2] == 0) {
			return 1;
		}
	}
	return 0;
}

/*
 * look for type to Anonymous type hash table
 */
static typeinfo_t*
find_anonymous_type(typ)
	typeinfo_t *typ;
{
	typeinfo_t *t;

	assert(typ != NULL);
	assert(typ->type_hash != 0);

	for (t = anonymous_types[typ->type_hash & (TYPE_HASH_MAX-1)]; t; t = t->type_anonymous_list) {
		if (equal_types(t, typ)) {
			return t;
		}
	}

	return NULL;
}

/*
 * add type to Anonymous type hash table
 */
static void
store_anonymous_type(typ)
	typeinfo_t *typ;
{
	int index;
	index = typ->type_hash & (TYPE_HASH_MAX-1);
	typ->type_anonymous_list = anonymous_types[index];
	anonymous_types[index] = typ;
}

/*
 * Define some builtin/intrinsic anonymous types
 */
void
init_anonymous_types()
{
	static int initialized = 0;
	typeinfo_t *typ;
	symbol_t *basetype;

	if (initialized) return;

	initialized = 1;

	typ = add_pointer_type(typeof_char());
	typ->_anonymous = 1;

	basetype = new_sym();
	basetype->intrinsic = 1;
	typ->type_base = basetype;

	store_anonymous_type(typ);

	basetype->sym_kind = type_symbol;
	basetype->sym_type = typ;

	basetype->sym_ident = new_node(_Ident, "%% builin *char %%");
	basetype->sym_ada_name = TYPE_NAMEOF_CHAR_POINTER;
	basetype->gened = 1;

	typ = add_pointer_type(typeof_void());

	basetype = new_sym();
	basetype->intrinsic = 1;
	typ->type_base = basetype;

	store_anonymous_type(typ);

	basetype->sym_kind = type_symbol;
	basetype->sym_type = typ;

	basetype->sym_ident = new_node(_Ident, "%% builin *void %%");
	basetype->sym_ada_name = "system.address";
	basetype->gened = 1;

	typ = add_pointer_type(add_function_type(typeof_int()));
	typ->_anonymous = 1;

	basetype = new_sym();
	basetype->intrinsic = 1;
	typ->type_base = basetype;

	store_anonymous_type(typ);

	basetype->sym_kind = type_symbol;
	basetype->sym_type = typ;

	basetype->sym_ident = new_node(_Ident, "%% int (*)() %%");
	basetype->sym_ada_name = TYPE_NAMEOF_FUNCTION_POINTER;
	basetype->gened = 1;

	anonymous_function_pointer = basetype;
}

/*
 * Here we go.  We may need to generate an anonymous type
 * for the input type.  An example is when "int *a;" is
 * encountered we'll need to gen "type a_int_t is ...".
 */
symbol_t*
get_anonymous_type(typ)
	typeinfo_t *typ;
{
	typeinfo_t *anonymous_type;
	symbol_t *basetype;
	ident_case_t icase;
	char buf[512];

	assert(typ != NULL);

	if (typ->type_next != NULL && typ->type_next->type_base == NULL) {
		if (is_function_pointer(typ)) {
			typ->type_base = anonymous_function_pointer;
			return typ->type_base;
		}
#if 0
		dump_type(typ);
#endif
	}

	assert(typ->type_next == NULL || typ->type_next->type_base != NULL);

	assert(typ->type_hash != 0);
	assert(typ->type_anonymous_list == NULL);

	anonymous_type = find_anonymous_type(typ);
	if (anonymous_type != NULL) {
		basetype = anonymous_type->type_base;
		assert(basetype != NULL);
		typ->type_base = basetype;
		return basetype;
	}

	if (is_access_to_record(typ)) {
		typeinfo_t *rtyp;

		rtyp = typ->type_next;
		assert(rtyp != NULL);

		basetype = rtyp->type_base;
		assert(basetype != NULL);
		assert(basetype->gened);

		/*
		 * I personally hate this type naming convention...
		 */
		if(rtyp->_constant) {
		    strcpy(buf, "ac_");
		} else {
			strcpy(buf, "a_");
		}
		if (basetype->sym_ada_name == NULL) {
			assert(basetype->sym_ident != NULL);
			assert(basetype->sym_ident->node_kind == _Ident);
			assert(basetype->sym_ident->node.id.name != NULL);
			basetype->sym_ada_name =
				ada_name(basetype->sym_ident->node.id.name, FILE_ORD(basetype->sym_def));
		}
		strcat(buf, basetype->sym_ada_name);
		icase = id_case(basetype->sym_ada_name);
		if (! ends_in_t(basetype->sym_ada_name)) {
			strcat(buf, "_t");
		}
		id_format(buf, icase);
	}
	else {
		sprintf(buf, "anonymous%d_t", next_anonymous_ord());
	}

	basetype = new_sym();
	typ->type_base = basetype;

	typ = copy_type(typ);
	typ->_anonymous = 1;
	store_anonymous_type(typ);

	basetype->sym_kind = type_symbol;
	basetype->sym_type = typ;
	basetype->sym_def = yypos;

	basetype->sym_ident = new_node(_Ident, new_string(buf));
	basetype->sym_ada_name =
		ada_name(basetype->sym_ident->node.id.name, FILE_ORD(yypos));
	
	return basetype;
}

